"""Constants."""

g = 9.81  # gravity acceleration (m/s^2)
surf_tens = 72.75e-3  # surface tension (N/m), ref. 20deg, from Vargaftik et al., 1983 - International tables of the
# surface tension of water, J.Phys.Chem.Ref.Data
density = 998.2  # water density (kg/m^3), ref. 20deg
